#pragma once

#include "common.h"
#include "Globals.h"
#include "ScopedPacketAccess.h"

/** PacketContext is an extremely important fundamental class.

	Resolution, worldspace location and dimensions, and convenient functions for converting between gridspace and worldspace are here.

	Many World Machine data packet types contain a PacketContext to describe their contents.

 ** @ingroup data
*/
class PacketContext {
public:
	enum SPACETYPE { WORLD, SPECIFIED, LOCAL };

	PacketContext();
	PacketContext(const PacketContext &other);
	/// Create a packet context describing a worldspace rectangle at a given resolution.
	PacketContext(const Rect4F &extents, const Coord &res, const SPACETYPE type = WORLD);
	PacketContext &operator= (const PacketContext &rhs);

	/// @name Set and Retrieve functions @{
	/// Set both the worldspace location and the maximum resolution of a side
	void set(const Rect4F &region, const int res);
	/// Set the resolution
	void	setRes(const int res);
	/// Set the worldspace render quadrant
	void	setRenderQuad(const Rect4F region);
	/// Set the vertical scale
	void	setVScale(const CoordF vscale);
	
	/// Set X/Y resolution to a specific value with no respect for aspect ratio, etc.
	/// Generally don't use this function unless you know what you're doing!
	/// Since this does not force the region aspect ratio to match the heightfield aspect ratio, you must
	/// use with care unless you are trying to do something funny.
	void	forceRes(const Coord res);

	Coord	getRes() const { return resolution; };
	int		getResMax() const { return (resolution.x > resolution.y) ? resolution.x : resolution.y; };
	inline int		w() const { return resolution.x; };
	inline int		h() const { return resolution.y; };
	CoordF	getRegionSize() const;

	/// Return the bottom left corner of the world extents
	CoordF		getRenderExtentOrigin() const;
	/// Retrieve the packet render extents
	RectF		getRenderExtent() const;
	const Rect4F &getRenderExtent4F() const;

	SPACETYPE	getType() const { return spacetype; };
	void setType(SPACETYPE st) { spacetype = st; };
	///@}

	/// @name Space Transform Functions @{
	/// Transform a world coordinate into a grid coordinate
	CoordF transformFromWorld(const CoordF &worldloc);
	Rect4F transformFromWorld(const Rect4F &worldrect);
	/// Transform an integer grid coordinate to a world coordinate
	CoordF transformToWorld(const Coord &gridcoordinate);
	/// Transform a floating point grid coordinate to a world coordinate
	CoordF transformToWorld(const CoordF &gridcoordinate);
	Rect4F transformToWorld(const Rect4F &gridrect);


	/// Clip a grid coordinate against the resolution bounds specified
	CoordF clipGridCoord(const CoordF &coord);
	Coord  clipGridCoord(const Coord &coord);
	/// Is the given grid coordinate in bounds?
	bool   testGridCoord(const Coord &coord);
	bool   testGridCoord(const CoordF &coord);

	/// Convert a raw grid height into a global height value
	float transformElevation(float elev) const { return vertical.x + vertical.y*elev; };
	/// Convert a world elevation in WM units to grid height
	float transformElevationFromWorld(float elev) const { return (elev-vertical.x) / vertical.y; };

	///@}

	/// Returns the detail scale, ie : pixels / worldspace unit.
	float getDScale() const { if (worldregion.width() > 0.0f) return resolution.x / worldregion.width(); return 0.0f; };

	/// Return the vertical scale (maximum different between smallest and largest elevations) of this packet
	float getVscale() const { return vertical.y; };
	/// Retrieve the minimum and maximum elevations.
	/// This is stored as a base,offset pair rather than an elevation pair.
	/// thus to get the upper bound, you will have to do vertical.x + vertical.y
	CoordF getVertical() const { return vertical; };

	/// Return the area of the packet in grid coordinates.
	size_t resarea() const { return size_t(resolution.x) * size_t(resolution.y); };
private:

	inline void createFastMul();

	Coord		resolution;		///< packet area primitives
	Rect4F		worldregion;	///< Region this packet encloses
	SPACETYPE	spacetype;		///< obeys worldspace?
	CoordF		vertical;		///< vertical base,scale pair
private:	
	CoordF		oor;			// one-over-resolution reciprocal for fast transforms
	CoordF		oos;			// one-over-worldsize reciprocal
};

#include "RescalerInfo.h"

/** BuildContext provides size and build state necessary for a device to build.
	@ingroup data, devices, building

The context information passed includes the preview state, render extents, memory conservation,
 as well as a reporting mechanism to pass progress reports back to the Builder.

BuildContext gets passed to your devices' Activate() functions.

BuildContext can be freely shallow-copied to allow for different interior arrangements
	
*/
class BuildContext : public PacketContext {
public:
	/** Construct a BuildContext.
	The build context is the primary interface point between your devices and the world machine core.
	As such, you must provide both the DeviceWorldBuilder responsible for the current build, and also
	the result set that you want to store the results of the build into.

	Without a builder, devices cannot report their build progress. Without build results, they cannot store their contents.
	*/
	BuildContext(BuildResultSet *_result, DeviceWorldBuilder *owner);
	BuildContext(const BuildContext &other);
	BuildContext &operator= (const BuildContext &rhs);

	void assignBuilder(DeviceWorldBuilder *_dwb) { builder = _dwb; };
	void assignResultSet(BuildResultSet *_result) { result_ptr = _result; };

	void SetWorldSize(RenderExtent &re, WorldCommonData &wpd);
	void SetWorldSize(Rect4F &region, int res, CoordF vscale, int tiled_blending_divide = 0);

	int getWorldResolution() { return world_resolution; };
	void forceWorldResolution(int w) { world_resolution = w; };
	int getTiledBlendAmount() { if (tiled) return tiled_blend_percent; else return 0; };

	// Build Progress Reporting operations
	virtual bool ReportDeviceProgress(Device *dev, int amnt);			 ///< Progress on a 0..100 scale
	virtual bool ReportDeviceProgress(Device *dev, int amnt, int max); ///< Progress normalize to 0..max

	// Store or retrieve data from the build context.
	BuildResultSet &results() { return *result_ptr; };
	DeviceResults &device(const Device *dev) const;

	bool setInput(Device *dev, int slot, PacketPointer data);
	bool setOutput(Device *dev, int slot, PacketPointer data);
	bool setParam(Device *dev, int slot, PacketPointer data);
	bool setMask(Device *dev, PacketPointer data);
	bool setCoordPacket(Device *dev, PacketPointer data);
	bool setDefaultOutput(Device *dev, PacketPointer data);

	PacketPointer getMask(Device *dev);
	PacketPointer getInput(Device *dev, int slot);
	PacketPointer getCoordPacket(Device *dev);
	PacketPointer getParamPacket( Device* dev, int port );
	ScopedPacketAccess viewDefaultOutput(Device *dev);

	void setBuilt(Device *dev);
	bool isBuilt(Device *dev);	

	bool keepChildResults();		///< if the current device should keep all internal results for inspection
	// Status variables set by the build process
	bool tiled;						///< flagged if this is a tiled build
	bool allow_device_threading;	///< if multithreading-capable devices should multithread
	bool pageable_results;			///< if true, build process will mark results as pageable
	bool export_outputs;			///< outputs should export their contents during this build
	bool conserve_memory;			///< use memory conservation; intermediate devices will not save build results
	bool canonical;					///< true if the build is "canonical"; that is, it accurately represents a build of the master render extent, and is not a temporary build (layout/etc)
	// Status variables set per-device
	bool terminal;					///< Set by builder on a per-device basis. true if this device is a terminal in the current build
	bool preview;					///< true if this build should be considered a preview. Preview builds do not participate in resolution-changing.
	WorldMachine::Rescale::RescaleInfo rescale_info;
private:
// Links to the two most important parts of a build context: the builder that is responsible for this context, and the result set of what the build produces
	DeviceWorldBuilder *builder;								///< The Builder that is orchestrating the build sequence, that we're working for
	BuildResultSet *result_ptr;
// Protect 
	/// as opposed to the current build resolution, the world resolution is the size specified by the user. This is needed
	/// for devices that use relative-scaled sizes
	int world_resolution;
	/// the tiled terrain blending percentage. This is needed for properly resizing fixed-size resolutions
	int tiled_blend_percent;
};