#ifndef _GEOMETRY_H
#define _GEOMETRY_H
#pragma once

/** @file Geometry.h

	Geometric helper classes and functions

*/

#include <math.h>
#include <vector>
#include "common.h"

class Geometry {

};

/** Templated 2D Vector class
 ** You can specialize on float, double, int...
 */
template < typename Storage >
class Vector2Typ {
public:
	Vector2Typ() {};
	Vector2Typ(Storage nx, Storage ny) { x = nx; y = ny; };
	inline void operator+=( const Vector2Typ &first) {
		x += first.x;
		y += first.y;
	};
	inline void operator-=( const Vector2Typ &first) {
		x -= first.x;
		y -= first.y;
	};
	inline void operator/=( const Storage divisor ) {
		x /= divisor;
		y /= divisor;
	};
	inline void operator/=( const Vector2Typ &first ) {
		x /= first.x;
		y /= first.y;
	};
	inline void operator*=( const Storage scalar ) {
		x *= scalar;
		y *= scalar;
	};
	inline void operator*=( const Vector2Typ &second ) {
		x *= second.x;
		y *= second.y;
	};

	inline Storage length() const  { return (Storage) sqrt((x*x)+(y*y)); };
	inline Storage length2() const  { return ((x*x)+(y*y)); };			// length squared
//	inline float lengthManhattan() const  { return (float) (x*x)+(y*y); };

	inline void normalize() { Storage size = length(); if (size > 0) { x /= size; y /= size; }};
	/// Returns unit vector in direction of this vector
	inline Vector2Typ direction() {	
		Storage size = length();
		if (size)
			return Vector2Typ(x / size, y / size);
		else
			return Vector2Typ(0,0);
	};

	Storage x,y;
};

template < typename T >
inline Vector2Typ<T> smallest( const Vector2Typ<T> &first, const Vector2Typ<T> &second ) {
	return Vector2Typ<T>(min(first.x, second.x), min(first.y, second.y));
};

template < typename T >
inline Vector2Typ<T> largest( const Vector2Typ<T> &first, const Vector2Typ<T> &second ) {
	return Vector2Typ<T>(max(first.x, second.x), max(first.y, second.y));
};


template < typename T >
inline Vector2Typ<T> operator+( const Vector2Typ<T> &first, const Vector2Typ<T> &second ) {
	Vector2Typ<T> result;
	result.x = first.x + second.x;
	result.y = first.y + second.y;
	return result;
};

template < typename T >
inline Vector2Typ<T> operator-( const Vector2Typ<T> &first, const Vector2Typ<T> &second ) {
	Vector2Typ<T> result;
	result.x = first.x - second.x;
	result.y = first.y - second.y;
	return result;
};

template < typename T >
inline Vector2Typ<T> operator*( const Vector2Typ<T> &first, const T scalar ) {
	Vector2Typ<T> result;
	result.x = first.x * scalar;
	result.y = first.y * scalar;
	return result;
};

template < typename T >
inline Vector2Typ<T> operator/( const Vector2Typ<T> &first, const Vector2Typ<T> &second ) {
	Vector2Typ<T> result;
	result.x = first.x / second.x;
	result.y = first.y / second.y;
	return result;
};

template < typename T >
inline Vector2Typ<T> operator/( const Vector2Typ<T> &first, const T scalar ) {
	Vector2Typ<T> result;
	result.x = first.x / scalar;
	result.y = first.y / scalar;
	return result;
};

template < typename T >
inline Vector2Typ<T> operator*( const Vector2Typ<T> &first, const Vector2Typ<T> &second ) {
	Vector2Typ<T> result;
	result.x = first.x * second.x;
	result.y = first.y * second.y;
	return result;
};

template < typename T >
inline T dot(const Vector2Typ<T> &first, const Vector2Typ<T> &second) {
	T dp = first.x*second.x + first.y*second.y;
	return dp;
};

template < typename T >
inline T RadBetween(const Vector2Typ<T> &first, const Vector2Typ<T> &second) {
// dotprod : ABcostheta = a.x*b.x + a.y*b.y
	T dp = first.x*second.x + first.y*second.y;
	dp = dp / (first.length() * second.length());
	return (T) acos(dp);
};

/// Convert a scalar value to a vector pair that matches the aspect of the supplied aspect vector.
/// The larger of the two valued aspect result returned will thus match the supplied scalar value.
/// (Or they will both match for square aspect)
template < typename VTYPE, typename T >
Vector2Typ<typename T> match_aspect( Vector2Typ<VTYPE> const&aspect, const T scalar ) 
{
	Vector2Typ<T> result;

	VTYPE maxV = max(aspect.x, aspect.y);

	result.x = static_cast<T> ( static_cast<VTYPE>(scalar) * aspect.x / maxV );
	result.y = static_cast<T> ( static_cast<VTYPE>(scalar) * aspect.y / maxV );

	return result;
};

// partially specialize for integer to get round-to-nearest behavior, which is frequently what we want as truncation causes weird shit with slightly-non-square aspects
template < typename VTYPE >
Vector2Typ<int> match_aspect( Vector2Typ<VTYPE> const&aspect, const int scalar ) 
{
	Vector2Typ<int> result;

	VTYPE maxV = max(aspect.x, aspect.y);

	result.x = static_cast<int> ( (static_cast<VTYPE>(scalar) * aspect.x / maxV) + 0.5f );
	result.y = static_cast<int> ( (static_cast<VTYPE>(scalar) * aspect.y / maxV) + 0.5f );

	return result;
};

/// float-based 3D vector
class Vector3 {
public:
	Vector3() {};
	Vector3(float _x, float _y, float _z) { x = _x; y = _y; z = _z;};
	float x,y,z;
	inline float length() { return (float) sqrt((x*x)+(y*y)+(z*z)); };
	inline float length2() { return ((x*x)+(y*y)+(z*z)); };			// length squared
	inline void normalize() { float size = length(); x /= size; y /= size; z /= size; };
	Vector3 operator+( Vector3 &second );
	Vector3 operator-( Vector3 &second );
	Vector3 operator*( float scalar );
	inline void operator+=( Vector3 &first) {
		x += first.x;
		y += first.y;
		z += first.z;
	};
	inline void operator-=( Vector3 &first) {
		x -= first.x;
		y -= first.y;
		z -= first.z;
	};
	inline void operator/=( float divisor ) {
		x /= divisor;
		y /= divisor;
		z /= divisor;
	};
	inline void operator*=( Vector3 &second ) {
		x *= second.x;
		y *= second.y;
		z *= second.z;
	};
	inline void operator*=( float scalar ) {
		x *= scalar;
		y *= scalar;
		z *= scalar;
	};
	inline float dot(Vector3 &other) {
		return other.x*x + other.y*y + other.z*z;
	};
};

inline Vector3 Vector3::operator+( Vector3 &other ) {
	Vector3 result;
	result.x = x + other.x;
	result.y = y + other.y;
	result.z = z + other.z;
	return result;
};

inline Vector3 Vector3::operator-( Vector3 &other ) {
	Vector3 result;
	result.x = x - other.x;
	result.y = y - other.y;
	result.z = z - other.z;
	return result;
};

inline Vector3 Vector3::operator*( float scalar ) {
	Vector3 result;
	result.x = x * scalar;
	result.y = y * scalar;
	result.z = z * scalar;
	return result;
};

/// integer 2D vector
struct Coord {
	Coord() {};
	Coord(int a,int b) { x = a; y = b; };
	Coord(Vector2Typ<int> other) { x = other.x; y = other.y; }; // constructor from generic int template
	friend Coord operator*( const Coord &first, const Coord &second );
	friend Coord operator+( const Coord &first, const Coord &second );
	friend Coord operator-( const Coord &first, const Coord &second );
	inline void operator+=( const Coord &first) {
		x += first.x;
		y += first.y;
	};
	inline void operator-=( const Coord &first) {
		x -= first.x;
		y -= first.y;
	};
	inline void operator/=( const int divisor ) {
		x /= divisor;
		y /= divisor;
	};
	inline void operator*=( const Coord &other ) {
		x *= other.x;
		y *= other.y;
	};
	inline void operator*=( const int scalar ) {
		x *= scalar;
		y *= scalar;
	};
	inline bool operator==( const Coord &other) const {
		return !operator!=(other);
	};
	inline bool operator!=( const Coord &other) const {
		if (other.x == x && other.y == y)
			return false;
		return true;
	};
	int x;
	int y;
};

inline Coord operator*( const Coord &first, const Coord &second ) {
	Coord result;
	result.x = first.x * second.x;
	result.y = first.y * second.y;
	return result;
};

inline Coord operator+( const Coord &first, const Coord &second ) {
	Coord result;
	result.x = first.x + second.x;
	result.y = first.y + second.y;
	return result;
};

inline Coord operator-( const Coord &first, const Coord &second ) {
	Coord result;
	result.x = first.x - second.x;
	result.y = first.y - second.y;
	return result;
};

inline Coord operator*( const Coord &first, const int scalar) {
	return Coord(first.x * scalar, first.y * scalar);
}

typedef Coord VectorI;
typedef VectorI Coord;

/// Int-based 2D Rectangle in point+size form
struct Rect {
	Rect() {};
	Rect(Coord p1, Coord p2);
	Rect(int _x, int _y, int _w, int _h) { x = _x; y = _y; width = _w; height = _h; };
	Rect( Rect &other) { x = other.x; y = other.y; width = other.width; height = other.height; };
	Rect &operator=( const Rect &other ) { x = other.x; y = other.y; width = other.width; height = other.height; return *this;};

	int x, y;
	int width, height;
};

/// Int based Rectangle (in 2-point form rather than point+size)
struct Rect4 {
	Rect4() {};
	Rect4(int _x0, int _y0, int _x1, int _y1) { x0 = _x0; y0 = _y0; x1 = _x1; y1 = _y1; };
	int x0, y0;
	int x1, y1;
};

struct Rect4F;

/// Float-based Rectangle in point+size form
struct RectF {
	RectF() {};
	RectF(const Rect4F &r);
	float x, y;
	float width, height;
};

/// Float-based Rectangle in 2-point form
struct Rect4F {
	Rect4F() { };
	Rect4F(float _x0, float _y0, float _x1, float _y1) { x0 = _x0; y0 = _y0; x1 = _x1; y1 = _y1; };
	Rect4F(const RectF &r);
	float x0, y0;
	float x1, y1;

	Rect toRect() const { Rect r(int(x0), int(y0), int(x1-x0), int(y1-y0)); return r; };

	void offset(CoordF &offset) { x0 += offset.x; x1 += offset.x; y0 += offset.y; y1 += offset.y; };

	float width() const { return x1 - x0; };
	float height() const { return y1 - y0; };

	Vector2 size() const { return Vector2(width(), height()); };
	Vector2 center() const { return Vector2(0.5f*(x0+x1), 0.5f*(y0+y1)); };

	float maxdimension() { Vector2 s = size(); return max(s.x, s.y); };

	void Validate(); // make sure x0, y0 are <= x1,y1
};


/// @name Conversion Functions
//@{
Rect4F RectF_to_Rect4F(RectF const&r);
RectF Rect4F_to_RectF(Rect4F const&r);
inline Coord f2i(CoordF const&f) { return Coord(int(f.x), int(f.y)); };
inline CoordF i2f(Coord const&i) { return CoordF(float(i.x), float(i.y)); };
//@}

/// 2D Polygon class
/// Consists of a series of Vector2 points, and includes basic geometric functions
class Polygon2 : public Geometry
{
public:
	Polygon2() { };
	Polygon2(Vector2 *ptlist, int ptsize);

	void addPoint(Vector2 _pt) { pts.push_back(_pt); };
	Rect4F bounds();

	bool IntersectPoint(const Vector2 &pt);	// True if point is inside Polygon
	bool IntersectRect(const Rect4F &rect);	// True if polygon and rectangle overlap

	bool isContainedBy(const Rect4F &container);// True if the polygon is completely within the bounds of the container

	void Rotate(const float angle);				// Rotates around origin
	void Translate(const Vector2 &offset);

	// Returns false if given location is out of bounds
	bool spanX(const float xval, CoordF &result);	// Retrieve polygon min/max span at given X value.
	bool spanY(const float yval, CoordF &result);	// Retrieve polygon min/max span at given Y value. 

	std::vector<Vector2> pts;
};

/// @name Intersection Functions
//@{
bool Contains_Rect4F(const Rect4F &outer, const Rect4F &inner);
bool Intersect_Rect(const Rect &a, const Rect &b);
bool Intersect_RectF(const RectF &a, const RectF &b);
bool Intersect_Rect4F(const Rect4F &a, const Rect4F &b);
bool Intersect_Point_Rect(const Coord &a, const Rect &b);
bool Intersect_Point_Rect4(const Coord &a, const Rect4 &b);
bool Intersect_Point_RectF(const CoordF &a,const RectF &b);
bool Intersect_Point_Rect4F(const CoordF &a, const Rect4F &b);
bool Intersect_Point_Poly(const CoordF &a, CoordF *bound, int numbound, Vector2 offset = Vector2(0,0));

bool Intersect_LinePoint(const Coord &a, const Coord &b, const Coord point, const float tolerance = 1.0f);
bool Intersect_LineRect4F(const CoordF &a, const CoordF &b, const Rect4F &rect);
CoordF Find_Intersect_Line_Line(const CoordF &a1,const CoordF &a2,const CoordF &b1,const CoordF &b2, float *u, float *v);
//@}

/// @name Geometric Functions
//@{
float distance(CoordF &a, CoordF &b);
float distance2(CoordF &a, CoordF &b);
Vector2 Rotate2(Vector2 pt, float angle);							// 2D rotation of a point in plane
Vector3 Rotate3(Vector3 pt, float xang, float yang, float zang);	// 3D rotation of a point in space
//@}

#endif