/** @file Globals.h
// All global data and functions that are accessible across all plugins/cores are here.
// Each plugin gets a pointer to these from the core when they register. The core itself
// has the actual data/implementations. This allows for better memory heap management across
// DLL boundaries, as well as eliminating the need for linking to an external core.dll.

*/

#pragma once
#ifndef _GLOBALS_H
#define _GLOBALS_H

#include "common.h"
#include "WMAllocator.h"
#include <vector>

enum DEVICESPACE { WORLDSPACE, LOCALSPACE };

enum ACTIVATE_RESULT { ACTIVATE_SUCCESS = 1, ACTIVATE_NOACTIVATE = 0, ACTIVATE_CURRENT = -1, ACTIVATE_DISABLED = -2, ACTIVATE_EXCEPTION = 2 };

/// Preview state, used to choose between preview, final, or temporary datastores.
enum PSTATE { P_ALL = -1, P_FINAL = 0, P_PREVIEW = 1, P_OVERSCAN = 2, PSTATE_SIZE = 3};

class OutputStore;
class Device;
class WMPacket;

/// @name Global Function Pointers
//@{
typedef Device* (DMakePtr)();
typedef void (DKillPtr)(Device* ptr);
typedef PacketPointer (PackMakePtr)();
typedef void (PackKillPtr)(PacketPointer ptr);
//@}

/// Device plugin management
/// The WM Core uses this structure to manage
/// the device (load, create, and destroy instances of the device class)
/// @ingroup devices application
struct DEV_DATA {		// Entry in device tag-library (1 entry per device derivative)
	DEV_DATA() { nametag[0] = NULL; maker = NULL; killer = NULL; nullterm = 0;};
	DEV_DATA(char snametag[4], DMakePtr smaker, DKillPtr skiller ) { maker = smaker; killer = skiller; strncpy(nametag, snametag, 4); };
	char nametag[4];	///< 4-byte device ID code
	char nullterm;
	DMakePtr *maker;	///< fpointer to create a device instance (in device's heap)
	DKillPtr *killer;	///< fpointer to destroy a device instance (in device's heap)
};

class HField;
class WMPacket;
class ColorInterface;
class CurvesInterface;

const int PROGRESS_SCALE = 100;

/** @name Convenience Wrappers
These macros provide convenient memory access
@ingroup memory
@relatesalso WMCoreInterface
*/
//@{
#define WMAlloc(x) WorldMachine::core->memory->Alloc(x)
#define WMFree(x) WorldMachine::core->memory->Free(x)
#define WMGetPacket(x) WorldMachine::core->memory->GetPacket(x)

/** Gnarly bit of template madness to allow us to easily create a new uninitialized packet of the same type as the given packet.
*/
template < class PACKET_TYPE >
boost::intrusive_ptr<PACKET_TYPE > WMPacketFrom(boost::intrusive_ptr<PACKET_TYPE > const&model) {
	PacketPointer result = WMGetPacket(model->getType());
	if (!result)
		return nullptr;

	assert(result->getType() == model->getType());
	PACKET_TYPE* ptr = (PACKET_TYPE*) result.get();
	return boost::intrusive_ptr<PACKET_TYPE >( ptr ); 
}

template < class PACKET_TYPE >
boost::intrusive_ptr<PACKET_TYPE > WMPacketFrom(PACKET_TYPE &model) {
	PacketPointer result = WMGetPacket(model.getType());
	if (!result)
		return nullptr;

	assert(result->getType() == model.getType());
	PACKET_TYPE* ptr = (PACKET_TYPE*) result.get();
	return boost::intrusive_ptr<PACKET_TYPE >( ptr ); 
}


// overrides for if you want to use new and delete out of the WM allocator. Allocs raw memory with WMAlloc
// then does placement new
template < class T>
T* WMNEW() {
	void *memory = WMAlloc(sizeof(T));
	T *result = new(memory) T;
	return result;
}

template < class T> 
void WMDELETE(T *ptr) {
	ptr->~T();
	WMFree(ptr);
}

//@}

class DeviceWorld;

const int GLOBAL_PATH_LENGTH = 2048;

enum READUNIT_TYPE { READUNIT_AUTO = -1, READUNIT_WM, READUNIT_KM, READUNIT_M };
enum WRITEUNIT_TYPE { WRITEUNIT_AUTO, WRITEUNIT_FORCE_M, WRITEUNIT_FORCE_KM, WRITEUNIT_FORCE_WM };

namespace WorldMachine {

class WMThreadPool;

/////////////////////////////////////////////////// NEW GLOBAL STRUC

/**@name Packet and Memory Management
 ** @ingroup application
 Memory allocation -- When inter operating with the WM Core, memory must be managed through this interface

 All data packets are created or deleted through this interface
 It also manages the optional packet caching and eviction model (if supported by the app core)

 Alloc is _guarenteed to return non-zero_ results; the core expects an exception thrown if memory cannot be allocated.
 */
class PacketCore {
public:
	/// Allocate a number of bytes of memory. If alloc_can_evict is true, this memory may trigger paging of vulnerable data to keep
	/// heap allocation size under a predefined limit.
	virtual void *Alloc(size_t size, bool alloc_can_evict = false) =0;
	virtual void Free(void *pointer) =0;
/// Allocate a new packet of a specific type
	virtual PacketPointer GetPacket(const int id) =0;
/// Please note that normally you would NEVER call FreePacket. Packets are automatically managed via smart pointer.
/// When they are detected to be unreferenced, they are deleted by internally calling this method.
	virtual void FreePacket(WMPacket * wmp) =0;

/// Return the number of active packets
	virtual int GetPacketCount() = 0;

/// Force app to page out all currently page-able datapackets to free up memory for new allocations
/// returns true if it was able to meet needs, or false if no more pagable packets are available.
	virtual bool PageOutPackets(size_t size_needed=0) =0;
};

/** @name Device Registration and Management
 ** @ingroup application
*/
class DeviceCore {
public:
/// Request a new device of the specified type
	virtual Device *GetDevice(const char* id) =0;
	/// Get a world-unique identifier for a device
	virtual int GetDeviceUniqueIdentifier(void) =0;

///@name Device Registration and Management
//@{
/// Register a device type with the core
	virtual bool RegisterDevice(DEV_DATA entry) =0;	
/// Retrieve entry for a given Device tag, if avail
	virtual DEV_DATA GetDeviceRegistry(const char* tag) =0;	
/// Clear all registered device types
    virtual void ClearRegistery() =0;
//@}
};


/** @name Application Interface. 
 ** Allows devices/core modules to make requests upon the application host.
 ** @ingroup application
 */
class AppCore {
public:
///@name Application paths to various folders that some devices might need to access for output,etc.
//@{
	virtual const char *getPathHome() =0;
	virtual const char *getPathData() =0;
	virtual const char *getPathDoc() =0;
	virtual const char *getPathMacro() =0;
	virtual const char *getPathPreset() =0;
	virtual const char *getPathTemp() =0;
//@}

///@name Unit conversion calls
//@{
	virtual float ReadUnit(const char *str, READUNIT_TYPE default_dimension) =0;
	virtual float ReadAltitudeUnit(const char *str) =0;
	virtual WMString WriteUnit(float wm_unit, WRITEUNIT_TYPE auto_metric_switching) =0;
	virtual WMString WriteAltitudeUnit(float wm_unit, WRITEUNIT_TYPE auto_metric_switching) =0;
//@}

	/// Get the size of the relationship between WM units and kilometers. This is usually 8km = 1WM
	virtual float getWM_KM_relation() = 0;

	/// Retrieve the preview size the Appcore is specifying
	virtual int getPreviewSize() = 0;

/// If the application core is purchased or not
	virtual int isCoreRegistered() =0;

/// If the application core is running in GUI mode or not
	virtual bool isCoreGUI() = 0;

/// Return the version version of the running version of WM
	virtual int getAppVersion() = 0;

/// If devices should expose their extra parameters
	virtual bool isExpandedPorts() = 0;

/// Returns true if the app is currently in a build process
	virtual bool isBuilding() = 0;

/// Opens up the given device in Layout:Overview:Device edit
	virtual void ShowDeviceInOverview(Device *device) = 0;

/** Post a request to run a function to the main thread of World Machine.
    This is an important ability for communication plugins as new data may come in from the other application
    and needs some way to signal World Machine to take some action in a thread-safe way.
    The function pointer specified will be run by the main program thread of WM and can thus call functions that
	would be otherwise unsafe from a different thread. 
*/
	virtual void RunFunction(Device *owner, RunFunctionInterface *postfunc) =0;
/// Clear any pending run requests for the given device. The framework will call this for a device when it is being
/// destroyed, but there may be additional cases where you want to call it yourself (to clear any previous requests for example)
	virtual void ClearRunFunction(Device *owner) =0;

/// Build the given world's preview
	virtual bool PreviewWorld(DeviceWorld *world) = 0;
/// Build the given world's final results!
	virtual bool BuildWorld(DeviceWorld *world) =0;
/// Ensure that the given device has up-to-date preview information
	virtual bool PreviewDevice(Device *dev) =0;

/// Retrieve the current heightfield colortable
	virtual void RetrieveColorTable(dword **colors, int *numcolors) =0;

/// Debugging (logging) functions
	virtual void Log(int channel, const char *message) =0;
};


/*** Provides a common set of UI dialogs, provided by the parent Application, for things such as creating curves, choosing colors, etc.
*/
class CommonUICore {
public:
	/// Display the application-defined generic data dialog for the given device
	virtual int GenericPropertyDialog(Device * dev);
	/// Display the application-defined generic name dialog for the given device
	virtual bool GenericNameDialog(Device *dev);
	/// Display the application-defined generic about dialog for the given device
	virtual void GenericAboutDialog(Device *dev);

	/// Display a message box to the user
	virtual void DisplayMessageBox(const char *txt);
	/// Request a choice from the user
	virtual unsigned int DisplayQueryBox(const char *txt, unsigned int flags);

	/// Retrieve a new Curves common dialog box
	virtual CurvesInterface *GetCurvesDisplay();
	/// Retrieve a new Color common dialog
	virtual ColorInterface *GetColorDisplay();
};

/** WM 2.3 and above Interface for the WMCore routines. These are implemented by the host and allow access to many different things.

  Each type of core functionality is implemented by the host and accessed through the data interface below. For the
  sake of organization, various functionality provided by the core is separated into different regimes.

  The application will set each functionality group interface to a valid interface.

  @ingroup memory application
*/
class WMCoreInterface {
public:
	AppCore *	app;
	PacketCore *memory;
	DeviceCore *device;
	CommonUICore *ui;
	WMThreadPool *threadpool;
};

// Each executable or DLL contains its own core pointer. If WM is being used across DLL boundaries, each plugin
// contains its own pointer to the core, but these pointers are all guarenteed to point to the same interface.
extern WMCoreInterface *core;

}
#endif