#pragma once
#include "common.h"
#include <vector>

class Device;
class WMPacket;
enum PSTATE;

const int PORT_NAME_LEN = 24;
/** Mechanism through which data moves into and out of the device.

Ports are the basic wiring glue of World Machine. Each port must have a parent device, and a device contains zero or more input Port, and zero or more OutputPort.
Each port can be connected to another port (or many in the case of OutputPorts). The interconnection of these ports configures the
device network.

Each port has an intrinsic type or types (those defined in WMPacket.h) associated with it. 

	@ingroup devices
	@ingroup data
*/
class Port {
public:
	Port();
	virtual ~Port(void);

	enum PORTFLAGS { NORMAL=0, SPECIAL=1, VARIABLE_RESOLUTION = 2};

/// @name Housekeeping functions
//@{
	void SetName(const char *_name);
	char *GetName() { return name; };

	Coord GetLoc(); 

	/// flag port with one or several of PORTFLAGS state.
	void SetFlag(int state) { flag = state; };
	int GetFlag() const { return flag; };

	void SetOwner(Device* _o) { owner = _o; };
	Device *Owner() const { return owner; };
//@}

///@name Graph/Link related members
//@{
/// Return the other port linked through this port
	virtual Port *Target() const { return target; };	// The other port
/// Return the other device that is linked through this port, or NULL if none
	virtual Device *Other() const { if (target) { return target->Owner(); }; return NULL; };

/// Connect this port to another
	virtual bool MakeLink(Port *other);
/// Break the connection this port has to parameter other. Includes route points.
	virtual bool BreakLink(Port *other = NULL);
/// Break any routes from this port to any other ports
	virtual bool BreakRoutes();
//@}

/// @name Type-related functions
/// See WMPacket.h for enumations
//@{
	bool AcceptsType(int type);				/// Tests to see if the port will accept this type
	int GetType() { return data_type; };
	void SetType(int _new) { data_type = _new; };
//@}

/// Send a notification back to our owner this port is either being connected or disconnected.
	void sendNotify( bool connecting );

protected:
	friend class OutputPort;
	virtual void SetTarget(Port *newlink, Port *oldlink = NULL);
	char name[PORT_NAME_LEN];
	int data_type;			///< what sorts of data can be taken at this port
	Port *target;
	Device *owner;
	int flag;				// is this port normal or special?
};

/** An OutputPort is an extension to the Port concept, allowing for a single (output) port to connect to multiple inputs.
	@ingroup devices
*/
class OutputPort : public Port {
public:
	OutputPort();
	virtual ~OutputPort(void);
//	OutputPort(const OutputPort &other);			// copy construction
//	OutputPort &operator=(const OutputPort &other);

// The (first) port we're connected to is default..
	virtual Port *Target() const;
	virtual Device *Other() const;

/// Return a specific output target
	Port *Target(int index) const;
/// Return a specific output target's owner
	Device *Other(int index) const;

	int numLinks() { return int(outlist.size()); };
	virtual bool MakeLink(Port *other);
	virtual bool BreakLink(Port *other = NULL);
	virtual bool BreakRoutes();
protected:
	virtual void SetTarget(Port *newlink, Port *oldlink = NULL);
private:
	std::vector<Port*> outlist;
};