#pragma once
#include "Context.h"

namespace WorldMachine {
	namespace Rescale {
		enum RESCALE_TYPE { DEFAULT, NEAREST, BILINEAR, BICUBIC_DEFAULT, BICUBIC_FAMILY, FRACTAL_SLOPE, FRACTAL_CURV };

		/** Rescale info block. This wraps the type of rescaling to be performed, and the optional roughness parameter */
		struct RescaleInfo {
			RescaleInfo(RESCALE_TYPE _t) { type = _t; roughness = 0.5f; };
			RescaleInfo(RESCALE_TYPE _t, float _s) { type = _t; roughness = _s; };
			RESCALE_TYPE type;
			float roughness;

			bool isWorldMethod() { return type == DEFAULT; };
			// Provides the world-default method of rescaling in case this rescaler is set to use the world default
			bool provideWorldMethod(RescaleInfo const&world) { if (type == DEFAULT) { type = world.type; roughness = world.roughness; return true; } return false; };

			bool supportsRoughness() { return type >= BICUBIC_FAMILY; };
		};

		// Additional information for fractal rescaling
		struct FractalRescaleInfo {
			FractalRescaleInfo() { worldloc = CoordF(0.0f, 0.0f); octaves = 1.0f; leadin = 0; };
			FractalRescaleInfo(PacketContext const&src, PacketContext const&dest) {
				octaves = max(0.0f, logf(dest.getDScale() / src.getDScale()) / logf(2.0f));
				leadin = 0;
			}
			void setGridLocation(CoordF const&wloc) { worldloc = wloc; };

			CoordF worldloc;
			float octaves;
			int leadin;
		};

	// Meta-programming support structures for RESCALE_TYPE
	// These traits tell the template code how many kernel values, etc to load up for the different resampling algorithms
		template < RESCALE_TYPE T>
		struct RescaleTypeTraits {
			// value 0 goes in spot 0, etc.
			enum KERNELRANGE { kernelMin = 0, kernelMax = 3 };
			// kernel offset adjustment: location of floor of coordinate in kernel relative to kernelMin.
			enum OFFSETADJ { zeroOffset = 1 };
		};

		// Nearest neighbor differs from the default:
		template <>
		struct RescaleTypeTraits<NEAREST> {
			// value0 goes in spot 1, only 2 values
			enum KERNELRANGE { kernelMin = 1, kernelMax = 2 };
			// kernel offset adjustment to get to spot 0.
			enum OFFSETADJ { zeroOffset = 1 };
		};

		// Bilinear differs from the default:
		template <>
		struct RescaleTypeTraits<BILINEAR> {
			// value0 goes in spot 1, only 2 values
			enum KERNELRANGE { kernelMin = 1, kernelMax = 2 };
			// kernel offset adjustment to get to spot 0.
			enum OFFSETADJ { zeroOffset = 1 };
		};

	}
}
