#pragma once

#include "ForwardDef.h"

/** ScopedPacketAccess makes sure that the packet you are attempting to work on is in-memory and not paged out to disk.
 ** 
 ** When data packets are "checked in" to a BuildResultSet, they can sometimes be paged out to disk for memory management purposes. By using a ScopedPacketLock, you can ensure that the
 ** data packet you are working on is not paged, and cannot be paged out while you have the lock.
 **
 ** Please note that if you are only working with packets provided to you in your Activate() function, or your own internal packets, you don't have to worry about this -- 
 ** the framework will automatically lock and unlock the packets it provides and takes from you.
 ** This class is most useful for Views inside the World Machine application, etc.
 **/
class ScopedPacketAccess
{
public:
	explicit ScopedPacketAccess(PacketPointer const& _ptr);
	ScopedPacketAccess(ScopedPacketAccess const&other);

	ScopedPacketAccess &operator=(PacketPointer const&rhs);
	ScopedPacketAccess &operator=(ScopedPacketAccess const&rhs);

	ScopedPacketAccess(ScopedPacketAccess &&rhs); // move constructor
	ScopedPacketAccess &operator=(ScopedPacketAccess &&rhs); // move assignment

	~ScopedPacketAccess(void);

	// validity of underlying pointer.
	operator bool() { return ptr != NULL; };

	PacketPointer const&get() const { return ptr; };
	void reset() { clear(); };
	WMPacket const&operator*() const { return *ptr; };
	WMPacket *operator->() const { return ptr.get(); };
private:
	void add();
	void clear();
	PacketPointer ptr;
};

