#pragma once

#include "common.h"
#include "globals.h"
#include "Context.h"

class HField;

struct TriangleIndexEntry {
	TriangleIndexEntry() {};
	TriangleIndexEntry(int v0, int v1, int v2) { v[0] = v0; v[1] = v1; v[2] = v2; };
	TriangleIndexEntry(int v0, int v1, int v2, int v3) { v[0] = v0; v[1] = v1; v[2] = v2; v[3] = v3; };
	int v[4];	// reserve space for 4 just in case we're using quads...
};

/// Simple triangle mesh class. You can store any amount of data at each vertex as this is a templated base class.
template < typename VertexType >
class TriangleMesh {
public:
	TriangleMesh();
	virtual ~TriangleMesh();

	// erase contents of the mesh
	virtual void clear();

	// copy contents from a different mesh
	void CopyFrom(const TriangleMesh &other);

	size_t getHeapUsage();	///< Return amount of WM heap memory used by this mesh


	bool isValid();

	// must be implemented by specialized type!
	void CalcBounds();
	Rect4F getXYBounds();
	CoordF getZBounds();

	// Vertex list
	VertexType *vertex_data;
	size_t num_vertex;
	// Face list
	TriangleIndexEntry *face_data;
	size_t num_face;

	Vector2 bx,by,bz;		///< bounding planes on X,Y,Z

	int primitive_type;		///< 0 = tris, 1 = quads
private:
	TriangleMesh &operator=(TriangleMesh const&other);
};

/** 3D Mesh primitive datatype. Specialized from TriangleMesh for vertices carrying only location data.
	@ingroup data
*/
class TerrainMesh : public TriangleMesh<Vector3> {
public:
	TerrainMesh();
	~TerrainMesh();

	void CopyFrom(const TerrainMesh &other);
	virtual void clear() override;

	enum CREATEFLAGS { CREATE_NORMALS = 1, CREATE_TEXCOORDS = 2, CREATE_FLIPEDGES = 4, CREATE_RECENTER = 8 };

/// Create a simple regular grid from a heightfield
	bool CreateRegularMesh(HFPointer hf, HFPointer mask, int create_flags = 0, Coord uvoffset = Coord(0,0));
/// THIS FUNCTION IS NOT YET IMPLEMENTED!!!
//	bool CreateOptimizedMesh(HFPointer hf, HFPointer mask, int quality, int create_flags = 0);

	void CalcBounds();
	Rect4F getXYBounds();
	// Return the true altitude bounds of this mesh
	CoordF getZBounds();

/// Experimental Export function: generate a heightfield from a mesh.
/// This has not been fully implemented. Do not use yet.
	HFPointer ExportHeightField(PacketContext &hfsize);

	bool usesQuads() const { return (primitive_type == 1); };

	int coord_system;		///< world scaling
	int primitive_type;		///< 0 = tris, 1 = quads
	bool use_tex_coords;	///< if the mesh contains texture coordinates
	bool use_normals;		///< if the mesh contains vertex normal data

	Vector2 *get_tex_coords() { return tex_coords; };
private:
	// additional per-vertex data associated with this mesh. These may not be present, and if not are null.
	Vector2 *tex_coords;
	Vector3 *normals;
};

class TerrainMeshImporter {
public:
	static bool InputOBJMesh(const char *filename, TerrainMesh &mesh);
};

class TerrainMeshExporter {
public:
	static bool OutputOBJMesh(const char *filename, TerrainMesh &mesh);
	static bool OutputLWOMesh(const char *filename, TerrainMesh &mesh);
};



// template implementation

template < typename VertexType >
TriangleMesh<VertexType>::TriangleMesh()
{
	vertex_data = NULL;
	face_data = NULL;
	num_vertex = 0;
	num_face = 0;
	bx.x = 10000.0f;
	bx.y =-10000.0f;
	by.x = 10000.0f;
	by.y =-10000.0f;
	bz.x = 10000.0f;
	bz.y =-10000.0f;
}

template < typename VertexType >
TriangleMesh<VertexType>::~TriangleMesh()
{
	TriangleMesh<VertexType>::clear();
}

template < typename VertexType >
void TriangleMesh<VertexType>::CopyFrom( const TriangleMesh &other )
{
	num_vertex = other.num_vertex;
	num_face = other.num_face;

	bx = other.bx;
	by = other.by;
	bz = other.bz;

	if (num_vertex > 0 && num_face > 0) {
		vertex_data = (Vector3*) WMAlloc(num_vertex*sizeof(Vector3));
		face_data = (TriangleIndexEntry*) WMAlloc(num_face*sizeof(TriangleIndexEntry));

		// transfer data 
		for (int i = 0; i < num_vertex; i++) {
			vertex_data[i] = other.vertex_data[i];
		}
		for (int i = 0; i < num_face; i++) {
			face_data[i] = other.face_data[i];
		}
	}
}

template < typename VertexType >
bool TriangleMesh<VertexType>::isValid()
{
	return vertex_data && face_data && (num_vertex > 0) && (num_face > 0);
}

template < typename VertexType >
size_t TriangleMesh<VertexType>::getHeapUsage()
{
	size_t size = (num_face * sizeof(TriangleIndexEntry)) + (num_vertex * sizeof(Vector3));
	return size;
}

template < typename VertexType >
void TriangleMesh<VertexType>::clear()
{
	if (vertex_data) {
		WMFree(vertex_data);
		vertex_data = NULL;
	}
	if (face_data) {
		WMFree(face_data);
		face_data = NULL;
	}
	num_vertex = 0;
	num_face = 0;
}
